@@ -186,7 +186,7 @@ cpl_boolean hdrl_rect_region_parameter_check(const hdrl_parameter * self)
 cpl_size hdrl_rect_region_get_llx(const hdrl_parameter * p) 
 {
     cpl_ensure(p, CPL_ERROR_NULL_INPUT, -1) ;
-    return ((hdrl_rect_region_parameter *)p)->llx;
+    return ((const hdrl_rect_region_parameter *)p)->llx;
 }
 
 /**
@@ -195,7 +195,7 @@ cpl_size hdrl_rect_region_get_llx(const hdrl_parameter * p)
 cpl_size hdrl_rect_region_get_lly(const hdrl_parameter * p) 
 {
     cpl_ensure(p, CPL_ERROR_NULL_INPUT, -1) ;
-    return ((hdrl_rect_region_parameter *)p)->lly;
+    return ((const hdrl_rect_region_parameter *)p)->lly;
 }
 
 /**
@@ -204,7 +204,7 @@ cpl_size hdrl_rect_region_get_lly(const hdrl_parameter * p)
 cpl_size hdrl_rect_region_get_urx(const hdrl_parameter * p) 
 {
     cpl_ensure(p, CPL_ERROR_NULL_INPUT, -1) ;
-    return ((hdrl_rect_region_parameter *)p)->urx;
+    return ((const hdrl_rect_region_parameter *)p)->urx;
 }
 
 /**
@@ -213,7 +213,7 @@ cpl_size hdrl_rect_region_get_urx(const hdrl_parameter * p)
 cpl_size hdrl_rect_region_get_ury(const hdrl_parameter * p) 
 {
     cpl_ensure(p, CPL_ERROR_NULL_INPUT, -1) ;
-    return ((hdrl_rect_region_parameter *)p)->ury;
+    return ((const hdrl_rect_region_parameter *)p)->ury;
 }
 
 /*----------------------------------------------------------------------------*/
@@ -231,7 +231,7 @@ cpl_error_code hdrl_rect_region_parameter_verify(
         const cpl_size              max_y)
 {
     const hdrl_rect_region_parameter * param_loc =
-        (hdrl_rect_region_parameter *)param ;
+        (const hdrl_rect_region_parameter *)param ;
 
     cpl_error_ensure(param != NULL, CPL_ERROR_NULL_INPUT,
             return CPL_ERROR_NULL_INPUT, "NULL Input Parameters");
@@ -257,7 +257,7 @@ cpl_error_code hdrl_rect_region_parameter_verify(
                          "urx %zu larger than maximum %zu",
                          (size_t)param_loc->urx, (size_t)max_x);
     if (max_y > 0)
-        cpl_error_ensure(param_loc->urx <= max_x, CPL_ERROR_ILLEGAL_INPUT,
+        cpl_error_ensure(param_loc->ury <= max_y, CPL_ERROR_ILLEGAL_INPUT,
                          return CPL_ERROR_ILLEGAL_INPUT,
                          "ury %zu larger than maximum %zu",
                          (size_t)param_loc->ury, (size_t)max_y);
@@ -281,14 +281,18 @@ cpl_error_code hdrl_rect_region_parameter_verify(
  */
 /* ---------------------------------------------------------------------------*/
 cpl_parameterlist * hdrl_rect_region_parameter_create_parlist(
-        const char              *   base_context,
-        const char              *   prefix,
-        const char              *   name_prefix,
-        const hdrl_parameter    *   defaults)
+        const char           *base_context,
+        const char           *prefix,
+        const char           *name_prefix,
+        const hdrl_parameter *defaults)
 {
     cpl_ensure(base_context && prefix && name_prefix && defaults,
                CPL_ERROR_NULL_INPUT, NULL);
-    cpl_parameterlist * parlist = cpl_parameterlist_new();
+
+    cpl_ensure(hdrl_rect_region_parameter_check(defaults),
+    		   CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
+
+    cpl_parameterlist *parlist = cpl_parameterlist_new();
 
     /* --prefix.llx */
     hdrl_setup_vparameter(parlist, prefix, ".", name_prefix,
@@ -325,7 +329,7 @@ cpl_parameterlist * hdrl_rect_region_parameter_create_parlist(
 /**
  @brief parse parameterlist for rectangle parameters
  @param parlist      parameter list to parse
- @param prefix       prefix of parameter
+ @param base_context prefix of parameter
  @param name_prefix  prefix of parameter name, may be empty string
  @see   hdrl_rect_get_parlist()
  @return A newly allocated hdrl_rect_region parameter or NULL in error case
@@ -337,17 +341,17 @@ cpl_parameterlist * hdrl_rect_region_parameter_create_parlist(
 /* ---------------------------------------------------------------------------*/
 hdrl_parameter * hdrl_rect_region_parameter_parse_parlist(
         const cpl_parameterlist *   parlist,
-        const char              *   prefix,
+        const char              *   base_context,
         const char              *   name_prefix)
 {
     cpl_size llx, lly, urx, ury ;
-    cpl_error_ensure(prefix && parlist, CPL_ERROR_NULL_INPUT,
+    cpl_error_ensure(base_context && parlist, CPL_ERROR_NULL_INPUT,
             return NULL, "NULL Input Parameters");
-    const char * sep = strlen(prefix) > 0 ? "." : "";
+    const char * sep = strlen(base_context) > 0 ? "." : "";
     const char * points[] = {"llx", "lly", "urx", "ury"};
     cpl_size * dest[] = {&llx, &lly, &urx, &ury};
     for (size_t i = 0; i < 4; i++) {
-        char * name = cpl_sprintf("%s%s%s%s", prefix, sep,
+        char * name = cpl_sprintf("%s%s%s%s", base_context, sep,
                                   name_prefix, points[i]);
         const cpl_parameter * par = cpl_parameterlist_find_const(parlist, name);
         *(dest[i]) = cpl_parameter_get_int(par);
@@ -357,13 +361,109 @@ hdrl_parameter * hdrl_rect_region_parameter_parse_parlist(
     if (cpl_error_get_code()) {
         cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                               "Error while parsing parameterlist "
-                              "with prefix %s", prefix);
+                              "with base_context %s", base_context);
         return NULL;
     }
 
     return hdrl_rect_region_parameter_create(llx, lly, urx, ury) ;
 }
 
+/*----------------------------------------------------------------------------*/
+/**
+  @brief    wrap negative or zero coordinates around full image size
+  @param    rect_region  rect region to wrap
+  @param    nx   image size in x, nx is added to entries < 1
+  @param    ny   image size in y, ny is added to entries < 1
+  @return  cpl_error_code
+
+  allows reverse indexing: 0 would be nx, -2 would be nx - 2 etc
+  Wrapping is based in FITS convention: 1 first pixel, nx last pixel inclusive
+ */
+/*----------------------------------------------------------------------------*/
+cpl_error_code hdrl_rect_region_fix_negatives(
+        hdrl_parameter    *     rect_region,
+        const cpl_size          nx,
+        const cpl_size          ny)
+{
+    hdrl_rect_region_parameter * rr_loc =
+        (hdrl_rect_region_parameter *)rect_region ;
+
+    cpl_error_ensure(rect_region != 0, CPL_ERROR_NULL_INPUT,
+            return CPL_ERROR_NULL_INPUT, "region input must not be NULL");
+    cpl_error_ensure(hdrl_rect_region_parameter_check(rect_region),
+            CPL_ERROR_ILLEGAL_INPUT, return CPL_ERROR_ILLEGAL_INPUT,
+            "Expected Rect Region parameter") ;
+
+    if (nx > 0 && rr_loc->llx < 1) rr_loc->llx += nx;
+    if (ny > 0 && rr_loc->lly < 1) rr_loc->lly += ny;
+    if (nx > 0 && rr_loc->urx < 1) rr_loc->urx += nx;
+    if (ny > 0 && rr_loc->ury < 1) rr_loc->ury += ny;
+
+    return hdrl_rect_region_parameter_verify(rect_region, nx, ny);
+}
+
+/* ---------------------------------------------------------------------------*/
+/**
+ @brief returns CPL_TRUE if x is strictly monotonic increasing
+ @param x      array to check for being strictly monotonic increasing
+ @param l      length of the array
+ @return CPL_TRUE if x is strictly monotonic increasing, CPL_FALSE otherwise
+ */
+/* ---------------------------------------------------------------------------*/
+cpl_boolean hdrl_is_strictly_monotonic_increasing(const double * x, cpl_size l){
+
+    if(x == NULL || l <= 1) return CPL_TRUE;
+
+    for(cpl_size i = 0; i < l - 1; i++){
+        if(x[i] >= x[i + 1]) return CPL_FALSE;
+    }
+    return CPL_TRUE;
+}
+
+/* ---------------------------------------------------------------------------*/
+/**
+ @brief sort in increasing or decreasing order x. Keep aligned with y1 and y2.
+ @param x               x  array
+ @param y1              y1 array
+ @param y2              y2 array
+ @param sample_len      length of the arrays
+ @param sort_decreasing if CPL_TRUE sort as decreasing, otherwise increasing
+
+ @return nothing
+ */
+/* ---------------------------------------------------------------------------*/
+void hdrl_sort_on_x(double * x, double * y1, double * y2,
+        const cpl_size sample_len, const cpl_boolean sort_decreasing){
+
+    cpl_propertylist * col_list = cpl_propertylist_new();
+
+    cpl_propertylist_append_bool(col_list, "x", sort_decreasing);
+
+    cpl_table * tb = cpl_table_new(sample_len);
+
+    cpl_table_wrap_double(tb, x,  "x");
+
+    if(y1)
+        cpl_table_wrap_double(tb, y1, "y1");
+
+    if(y2)
+        cpl_table_wrap_double(tb, y2, "y2");
+
+    cpl_table_sort(tb, col_list);
+
+    cpl_table_unwrap(tb, "x");
+
+    if(y1)
+        cpl_table_unwrap(tb, "y1");
+
+    if(y2)
+        cpl_table_unwrap(tb, "y2");
+
+    cpl_table_delete(tb);
+    cpl_propertylist_delete(col_list);
+}
+
+
 /* ---------------------------------------------------------------------------*/
 /**
  * @brief join strings together
@@ -404,40 +504,6 @@ char * hdrl_join_string(const char * sep_, int n, ...)
     return res;
 }
 
-/*----------------------------------------------------------------------------*/
-/**
-  @brief    wrap negative or zero coordinates around full image size
-  @param    rect_region  rect region to wrap
-  @param    nx   image size in x, nx is added to entries < 1
-  @param    ny   image size in y, ny is added to entries < 1
-  @return  cpl_error_code
-
-  allows reverse indexing: 0 would be nx, -2 would be nx - 2 etc
-  Wrapping is based in FITS convention: 1 first pixel, nx last pixel inclusive
- */
-/*----------------------------------------------------------------------------*/
-cpl_error_code hdrl_rect_region_fix_negatives(
-        hdrl_parameter    *     rect_region,
-        const cpl_size          nx,
-        const cpl_size          ny)
-{
-    hdrl_rect_region_parameter * rr_loc =
-        (hdrl_rect_region_parameter *)rect_region ;
-
-    cpl_error_ensure(rect_region != 0, CPL_ERROR_NULL_INPUT,
-            return CPL_ERROR_NULL_INPUT, "region input must not be NULL");
-    cpl_error_ensure(hdrl_rect_region_parameter_check(rect_region),
-            CPL_ERROR_ILLEGAL_INPUT, return CPL_ERROR_ILLEGAL_INPUT,
-            "Expected Rect Region parameter") ;
-
-    if (nx > 0 && rr_loc->llx < 1) rr_loc->llx += nx;
-    if (ny > 0 && rr_loc->lly < 1) rr_loc->lly += ny;
-    if (nx > 0 && rr_loc->urx < 1) rr_loc->urx += nx;
-    if (ny > 0 && rr_loc->ury < 1) rr_loc->ury += ny;
-
-    return hdrl_rect_region_parameter_verify(rect_region, nx, ny);
-}
-
 
 /** @cond EXPERIMENTAL */
 
@@ -569,11 +635,11 @@ char * hdrl_get_cwd(void)
  */
 /* ---------------------------------------------------------------------------*/
 cpl_error_code
-hdrl_normalize_imagelist_by_vector(const cpl_vector * scale,
-                                   const cpl_vector * scale_e,
-                                   const hdrl_scale_type scale_type,
-                                   cpl_imagelist * data,
-                                   cpl_imagelist * errors)
+hdrl_normalize_imagelist_by_vector(const cpl_vector      * scale,
+                                   const cpl_vector      * scale_e,
+                                   const hdrl_scale_type   scale_type,
+                                   cpl_imagelist         * data,
+                                   cpl_imagelist         * errors)
 {
     cpl_ensure_code(scale, CPL_ERROR_NULL_INPUT);
     cpl_ensure_code(scale_e, CPL_ERROR_NULL_INPUT);
@@ -751,8 +817,11 @@ cpl_vector * hdrl_image_to_vector(
     cpl_vector * vec_source = NULL;
     /* only cast if required, extra copying of non double data is still
      * faster than a loop with cpl_image_get */
+
+    CPL_DIAG_PRAGMA_PUSH_IGN(-Wcast-qual);
     cpl_image * d_img = cpl_image_get_type(source) == CPL_TYPE_DOUBLE ?
         (cpl_image *)source : cpl_image_cast(source, CPL_TYPE_DOUBLE);
+    CPL_DIAG_PRAGMA_POP;
 
     const cpl_size naxis1 = cpl_image_get_size_x(source);
     const cpl_size naxis2 = cpl_image_get_size_y(source);
@@ -919,14 +988,13 @@ void hdrl_cplvector_delete_to_cache(hdrl_vector_cache * cache, cpl_vector * v)
     if (!v) {
         return;
     }
-    if (!cache) {
-        return cpl_vector_delete(v);
-    }
-    cpl_size sz = cpl_vector_get_size(v);
-    if (sz < cache->nbuckets) {
-        if (cache->buckets[sz].available < cache->buckets[sz].nspace) {
-            cache->buckets[sz].ptrs[(cache->buckets[sz].available)++] = v;
-            return;
+    if (cache) {
+        const cpl_size sz = cpl_vector_get_size(v);
+        if (sz < cache->nbuckets) {
+            if (cache->buckets[sz].available < cache->buckets[sz].nspace) {
+                cache->buckets[sz].ptrs[(cache->buckets[sz].available)++] = v;
+                return;
+            }
         }
     }
     cpl_vector_delete(v);
@@ -944,11 +1012,9 @@ static cpl_vector * imagelist_to_vector(const cpl_imagelist * list,
                                         const cpl_binary ** maskbuf, hdrl_vector_cache * cache)
 {
     const long nz  = list ? cpl_imagelist_get_size(list) : -1;
-    double * restrict ddata = NULL;
     unsigned long j = 0;
-
-    cpl_vector * vec = hdrl_cplvector_new_from_cache(cache, nz);
-    ddata = cpl_vector_get_data(vec);
+    cpl_vector * vec        = hdrl_cplvector_new_from_cache(cache, nz);
+    double * restrict ddata = cpl_vector_get_data(vec);
 
     if (imgdatabuf && maskbuf) {
         for (long k = 0; k < nz; k++) {
@@ -1045,7 +1111,7 @@ cpl_error_code hdrl_imagelist_to_vector_row(const cpl_imagelist * list,
 {
     long nx;
     int isdouble;
-    const long nz  = list ? (long)cpl_imagelist_get_size(list) : -1;
+    const long nz  = list ? (long)cpl_imagelist_get_size(list) : 0;
     cpl_ensure_code(list != NULL, CPL_ERROR_NULL_INPUT);
     cpl_ensure_code(nz > 0, CPL_ERROR_ILLEGAL_INPUT);
     cpl_ensure_code(y > 0, CPL_ERROR_ACCESS_OUT_OF_RANGE);
@@ -1182,6 +1248,7 @@ hdrl_medianfilter_image_grid(const cpl_image * ima, cpl_matrix * x, cpl_matrix *
     }
     return ima_local;
 }
+
 /* ---------------------------------------------------------------------------*/
 /**
  * @brief create linear space row vector
@@ -1334,6 +1401,9 @@ image_const_row_view_create(const cpl_image * img,
     const char * d = cpl_image_get_data_const(img);
     size_t offset = (ly - 1) * nx;
     cpl_size nny = uy - ly + 1;
+
+    CPL_DIAG_PRAGMA_PUSH_IGN(-Wcast-qual);
+
     cpl_image * wimg = cpl_image_wrap(nx, nny, cpl_image_get_type(img),
                                      (char*)d + offset * dsz);
 
@@ -1344,14 +1414,20 @@ image_const_row_view_create(const cpl_image * img,
         cpl_mask_delete(hcpl_image_set_bpm(wimg, mask));
     }
 
+    CPL_DIAG_PRAGMA_POP;
+
     return wimg;
 }
 
 static void
 image_const_row_view_delete(const cpl_image * img)
 {
+    CPL_DIAG_PRAGMA_PUSH_IGN(-Wcast-qual);
+
     cpl_mask_unwrap(cpl_image_unset_bpm((cpl_image*)(img)));
     cpl_image_unwrap((cpl_image *)img);
+
+    CPL_DIAG_PRAGMA_POP;
 }
 
 /* ---------------------------------------------------------------------------*/
@@ -1393,14 +1469,15 @@ hdrl_parallel_filter_image(const cpl_image * img,
         ky = cpl_mask_get_size_y(mask);
         kx = cpl_mask_get_size_x(mask);
     }
+    cpl_ensure(ky % 2 == 1, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
+    cpl_ensure(ny >= ky, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
+    cpl_ensure(nx >= kx, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
+
     intptr_t hk = (ky / 2);
     cpl_image * res = cpl_image_new(nx, ny, cpl_image_get_type(img));
     /* make sure image has a bpm to avoid creation races later */
     cpl_image_get_bpm(res);
 
-    cpl_ensure(ky % 2 == 1, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
-    cpl_ensure(ny >= ky, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
-    cpl_ensure(nx >= kx, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
     /* filter first half-kernel rows (needs full kernel image) */
     if (hk) {
         const cpl_image * slice = image_const_row_view_create(img, 1, ky);
@@ -1461,6 +1538,78 @@ HDRL_OMP(omp parallel for lastprivate(y) if (ny > s + ky))
     return res;
 }
 
+
+/**
+ * @brief
+ *    Convert between physical and world coordinates using multiple threads
+ *
+ * @param wcs       The input cpl_wcs structure
+ * @param from      The input coordinate matrix
+ * @param to        The output coordinate matrix
+ * @param status    The output status array
+ * @param transform The transformation mode
+ *
+ * @return  An appropriate error code
+ *
+ * @see cpl_wcs_convert()
+ *
+ * This function is a wrapper around cpl_wcs_convert using OpenMP for
+ * parallelization, see the cpl documentation for details on the functionality.
+ */
+    /* ---------------------------------------------------------------------------*/
+cpl_error_code
+hdrl_wcs_convert(const cpl_wcs *wcs, const cpl_matrix *from,
+                 cpl_matrix **to, cpl_array **status,
+                 cpl_wcs_trans_mode transform)
+{
+    size_t nr = cpl_matrix_get_nrow(from);
+    size_t nc = cpl_matrix_get_ncol(from);
+    const size_t s = 4000;
+    size_t i;
+    int * dstatus;
+    cpl_ensure_code(to, CPL_ERROR_NULL_INPUT);
+    cpl_ensure_code(status, CPL_ERROR_NULL_INPUT);
+    cpl_ensure_code(wcs, CPL_ERROR_NULL_INPUT);
+    cpl_ensure_code(from, CPL_ERROR_NULL_INPUT);
+
+    *status = cpl_array_new(nr, CPL_TYPE_INT);
+    cpl_ensure_code(*status, CPL_ERROR_NULL_INPUT);
+    dstatus = cpl_array_get_data_int(*status);
+    *to = cpl_matrix_new(nr, nc);
+    cpl_error_code err = CPL_ERROR_NONE;
+
+HDRL_OMP(omp parallel for if (nr > s))
+    for (i = 0; i < nr; i+=s) {
+        cpl_matrix * lfrom = cpl_matrix_extract(from, i, 0, 1, 1, CPL_MIN(s, nr - i), nc);
+        cpl_matrix * lto = NULL;
+        cpl_array * lstatus = NULL;
+        cpl_error_code lerr = cpl_wcs_convert(wcs, lfrom, &lto, &lstatus, transform);
+        if (lto != NULL) {
+            cpl_matrix_copy(*to, lto, i, 0);
+        }
+        if (lstatus != NULL) {
+            memcpy(&dstatus[i], cpl_array_get_data_int(lstatus),
+                   cpl_array_get_size(lstatus) * sizeof(*dstatus));
+        }
+        cpl_array_delete(lstatus);
+        cpl_matrix_delete(lfrom);
+        cpl_matrix_delete(lto);
+        if (lerr != CPL_ERROR_NONE) {
+HDRL_OMP(omp critical(hdrlwcserror))
+            err = lerr;
+        }
+    }
+
+    if (err == CPL_ERROR_UNSUPPORTED_MODE) {
+        cpl_matrix_delete(*to);
+        *to = NULL;
+        cpl_array_delete(*status);
+        *status = NULL;
+    }
+
+    return cpl_error_set(cpl_func, err);
+}
+
 /* cpl_image_set_bpm from cpl 6.4 which we can't require yet for VLTSW 2014 */
 struct _cpl_image_ {
     /* Size of the image in x and y */
@@ -1489,6 +1638,21 @@ cpl_mask * hcpl_image_set_bpm(cpl_image * self, cpl_mask * bpm)
 #endif
 }
 
+double hcpl_vector_get_mad_window(cpl_vector * vec,
+                                  cpl_size llx,
+                                  cpl_size urx,
+                                  double * sigma)
+{
+    /* TODO use cpl directly when PIPE-4792 is fixed */
+    struct _cpl_image_ img;
+    img.pixels = cpl_vector_get_data(vec);
+    img.nx = cpl_vector_get_size(vec);
+    img.ny = 1;
+    img.bpm = NULL;
+    img.type = CPL_TYPE_DOUBLE;
+    return cpl_image_get_mad_window(&img, llx, 1, urx, 1, sigma);
+}
+
 double hcpl_gaussian_eval_2d(const cpl_array * self, double x, double y)
 {
 #if CPL_VERSION_CODE < CPL_VERSION(6, 4, 0)
@@ -1531,4 +1695,310 @@ double hcpl_gaussian_eval_2d(const cpl_array * self, double x, double y)
 
 /** @endcond */
 
+
+/*----------------------------------------------------------------------------*/
+/**
+ *
+ * @brief   Compute the effective airmass of an observation. Takes in count the
+ * error propagation if you enter the relative error of the input parameters
+ * in a hdrl_value structure {data,error}
+ *
+ * The function calculates the average airmass for the line of sight given by the
+ * right ascension (ra) and the declination (dec). The latitude (geolat) in
+ * degrees of the observatory site and the local siderial time (lst) at
+ * observation start has to be given, as well as the duration of the observation,
+ * i.e. the exposure time (exptime). If the exposure time is zero then only one
+ * value of airmass is computed, instead of weighting beginning, middle, and
+ * end of exposure according to Stetson (Stetson P., 1987, PASP 99, 191).
+ *
+ * @param   aRA          right ascension in degrees
+ * @param   aDEC         declination in degrees
+ * @param   aLST         local sideral time (s.) elapsed since siderial midnight
+ * @param   aExptime     integration time in seconds
+ * @param   aLatitude    latitude of the observatory site in degrees
+ * @param   type         kind of approximation (see hdrl_airmass_approx enumtype)
+ *
+ * @return  The computed average airmass or {-1,0.} on error.
+ *
+ * @note This function can be calcule different kinds of approximation specified
+ * in the hdrl_airmass_approx enum type, explain as follow. 1) The formula given
+ * by Hardie (Hardie 1962, In: "Astronomical Techniques", ed. Hiltner, p. 184)
+ * to compute the airmass as a function of zenith distance. 2) The formula of
+ * Young and Irvine (Young A. T., Irvine W. M., 1967, Astron. J. 72, 945).
+ * The range of trustworthy airmass outputs is only between 1 and 4. 3)
+ * The formula of Young (Young A. T., 1994 ApOpt, 33, 1108).
+ *
+ * @note You can to find in <a href="http://en.wikipedia.org/w/index.php
+ * ?title=Airmass&oldid=358226579#Interpolative_formulas">Interpolative formulas</a>
+ * an interesting collection of approximations.
+ *
+ */
+/*----------------------------------------------------------------------------*/
+hdrl_value hdrl_utils_airmass(
+    hdrl_value aRA, hdrl_value aDEC, hdrl_value aLST,
+	hdrl_value aExptime, hdrl_value aLatitude,
+	hdrl_airmass_approx type)
+{
+
+	hdrl_value retErr = {-1.,0.};
+    cpl_ensure(   aRA.data       >=   0. && aRA.data       <  360.   && aRA.error       >= 0.
+     		   && aDEC.data      >= -90. && aDEC.data      <= 90.    && aDEC.error      >= 0.
+			   && aLST.data      >=   0. && aLST.data      <  86400. && aLST.error      >= 0.
+			   && aExptime.data  >=   0.                             && aExptime.error  >= 0.
+			   && aLatitude.data >= -90. && aLatitude.data <= 90.    && aLatitude.error >= 0.
+			   && (type == HDRL_AIRMASS_APPROX_HARDIE || type == HDRL_AIRMASS_APPROX_YOUNG_IRVINE || type == HDRL_AIRMASS_APPROX_YOUNG),
+			   CPL_ERROR_ILLEGAL_INPUT,
+			   retErr);
+
+    /* Compute hour angle of the observation in degrees. */
+    hdrl_value HA = {aLST.data * 15. / 3600. - aRA.data,
+    				 aLST.error * fabs(15. / 3600.) + aRA.error * fabs(-1.)};
+
+    /* Range adjustments.Angle between line of sight and meridian is needed. */
+    if (HA.data < -180.) HA.data += 360.;
+    if (HA.data >  180.) HA.data -= 360.;
+
+    /* Convert angles from degrees to radians. */
+
+    hdrl_value delta     = {aDEC.data       *      CPL_MATH_RAD_DEG,
+                            aDEC.error      * fabs(CPL_MATH_RAD_DEG)};
+
+    hdrl_value latitude  = {aLatitude.data  *      CPL_MATH_RAD_DEG,
+                            aLatitude.error * fabs(CPL_MATH_RAD_DEG)};
+
+    hdrl_value hourangle = {HA.data         *      CPL_MATH_RAD_DEG,
+                            HA.error        * fabs(CPL_MATH_RAD_DEG)};
+
+
+    /* Calculate airmass of the observation using the approximation given    *
+     * by Young (1994). For non-zero exposure times these airmass values are *
+     * averaged using the weights given by Stetson.                          */
+
+    hdrl_value cosz = hdrl_get_zenith_distance(hourangle, delta, latitude);
+    hdrl_value z    = {0., cosz.error};
+    double zlimit   = 80.;
+
+	if (type == HDRL_AIRMASS_APPROX_HARDIE) {
+
+		z.data  = acos(cosz.data) * CPL_MATH_DEG_RAD;
+		z.error = cosz.error * fabs(-CPL_MATH_DEG_RAD / sqrt(1. + pow(cosz.data, 2)));
+
+		cpl_error_ensure(z.data <= zlimit, CPL_ERROR_ILLEGAL_OUTPUT, return retErr,
+			"Zenith angle %f+-[%f] > %f!", z.data, z.error, zlimit);
+	}
+
+	cpl_error_ensure(
+		cosz.data != 0. && fabs(1./cosz.data) >= FLT_EPSILON && acos(cosz.data) <= CPL_MATH_PI_2,
+		CPL_ERROR_ILLEGAL_OUTPUT,
+		return retErr,
+		"Airmass computation unsuccessful. Object is below the horizon at start (z = %f). Return the error",
+		cosz.error * fabs(-CPL_MATH_DEG_RAD / sqrt(1. + pow(cosz.data, 2))) );
+
+    hdrl_value airmass     = {0.,0.};
+	hdrl_value secansZdist = {1. / cosz.data, cosz.error * fabs(-1. / pow(cosz.data, 2))};
+
+    switch(type){
+		case HDRL_AIRMASS_APPROX_HARDIE:
+			airmass = hdrl_get_airmass_hardie(secansZdist);
+			break;
+        case HDRL_AIRMASS_APPROX_YOUNG_IRVINE:
+        	airmass = hdrl_get_airmass_youngirvine(secansZdist);
+        	break;
+        case HDRL_AIRMASS_APPROX_YOUNG:
+        	airmass = hdrl_get_airmass_young(cosz);
+        	break;
+    }
+
+    /* if the exposure time is larger than zero, weight in airmass   *
+     * at mid and end exposure according to Stetson's weights (which *
+     * are also used in IRAF/astcalc for the airmass function        */
+	if (aExptime.data > 0.) {
+
+		const double weights[]  = {1./6., 2./3., 1./6.};
+		const cpl_size nweights = sizeof(weights) / sizeof(double);
+
+		hdrl_value timeStep = {aExptime.data / (nweights - 1.) * 15. / 3600. * CPL_MATH_RAD_DEG, 0.};
+		timeStep.error      =  aExptime.error * fabs( 1. / (nweights - 1.) * 15. / 3600. * CPL_MATH_RAD_DEG);
+
+		airmass.data  *= weights[0];
+		airmass.error *= fabs(weights[0]);
+
+		for (cpl_size i = 1; i < nweights; i++) {
+
+			hdrl_value aux_hourangle = { hourangle.data  + i * timeStep.data,
+										 hourangle.error + i * timeStep.error };
+
+			cosz = hdrl_get_zenith_distance(aux_hourangle, delta, latitude);
+
+			if (type == HDRL_AIRMASS_APPROX_HARDIE) {
+
+				z = (hdrl_value){acos(cosz.data) * CPL_MATH_DEG_RAD, 0.};
+				z.error = cosz.error * fabs(-CPL_MATH_DEG_RAD / sqrt(1. + pow(cosz.data, 2)));
+
+				cpl_error_ensure(z.data <= zlimit, CPL_ERROR_ILLEGAL_OUTPUT, return retErr,
+					"Zenith angle %f+-[%f] > %f!", z.data, z.error, zlimit);
+			}
+
+			cpl_error_ensure(
+				cosz.data != 0. && fabs(1./cosz.data) >= FLT_EPSILON && acos(cosz.data) <= CPL_MATH_PI_2,
+				CPL_ERROR_ILLEGAL_OUTPUT,
+				return retErr,
+				"timeStep. Object is below the horizon at %s exposure (z=%f).",
+				i==1 ? "mid. Return the error" : "end. Return the error",
+				cosz.error * fabs(-CPL_MATH_DEG_RAD / sqrt(1. + pow(cosz.data, 2))) );
+
+			hdrl_value weight = {0., 0.};
+			secansZdist = (hdrl_value){1. / cosz.data, cosz.error * fabs(-1. / pow(cosz.data, 2))};
+
+			switch(type){
+				case HDRL_AIRMASS_APPROX_HARDIE:
+					weight = hdrl_get_airmass_hardie(secansZdist);
+					break;
+				case HDRL_AIRMASS_APPROX_YOUNG_IRVINE:
+					weight = hdrl_get_airmass_youngirvine(secansZdist);
+					break;
+				case HDRL_AIRMASS_APPROX_YOUNG:
+					weight = hdrl_get_airmass_young(cosz);
+					break;
+			 }
+			 airmass.data  += weights[i] *weight.data;
+			 airmass.error += weights[i] *weight.error;
+		}
+	}
+
+	if (type == HDRL_AIRMASS_APPROX_YOUNG_IRVINE) {
+
+		/* Accuracy limit for airmass approximation of Young & Irvine */
+		const double airmasslimit = 4.;
+
+		cpl_error_ensure(airmass.data <= airmasslimit, CPL_ERROR_ILLEGAL_OUTPUT,
+			return retErr, "Airmass larger than %f", airmasslimit);
+	}
+
+	return airmass;
+}
+
+/*----------------------------------------------------------------------------*/
+/**
+  @private
+  @brief   Compute the zenith distance for an observation.
+  @param   aHourAngle   Hour angle in radians.
+  @param   aDelta       Declination in radians.
+  @param   aLatitude    Latitude of the observatory in radians.
+  @return  cos(z) on success or 0. on error.
+
+  The function computes the cosine of the zenith distance for an observation
+  taken at an angle aHourAngle from the meridian, which can take values in the
+  range extending from @f$-\pi@f$ to @f$\pi@f$, and the declination aDelta with
+  possible values between @f$-0.5\pi@f$ and @f$0.5\pi@f$. The latitude
+  aLatitude of the observing site may take values in the range @f$0@f$ to
+  @f$2\pi@f$.
+ */
+/*----------------------------------------------------------------------------*/
+hdrl_value hdrl_get_zenith_distance(
+	hdrl_value aHourAngle, hdrl_value aDelta, hdrl_value aLatitude)
+{
+
+	hdrl_value p0 = {sin(aLatitude.data) * sin(aDelta.data),
+	                  aLatitude.error * fabs( cos(aLatitude.data) * sin(aDelta.data)
+			        + aDelta.error    * fabs( sin(aLatitude.data) * cos(aDelta.data)))};
+
+	hdrl_value p1 = {cos(aLatitude.data) * cos(aDelta.data),
+	                  aLatitude.error * fabs(-sin(aLatitude.data) * cos(aDelta.data)
+			        + aDelta.error    * fabs(-cos(aLatitude.data) * sin(aDelta.data)))};
+
+	hdrl_value z  = {p0.data + cos(aHourAngle.data) * p1.data,
+	                  p0.error         * fabs(1.)
+	                + aHourAngle.error * fabs(-sin(aHourAngle.data) * p1.data)
+	                + p1.error         * fabs( cos(aHourAngle.data))};
+
+	return fabs(z.data) < FLT_EPSILON ? (hdrl_value){0.,0.} : z;
+}
+
+/*----------------------------------------------------------------------------*/
+/**
+  @private
+  @brief   Compute airmass with Hardie (1962) approximation.
+  @param   aSecZ   The secans of the zenith distance.
+  @return  The function returns the airmass.
+
+  The function uses the approximation given by Hardie (Hardie 1962, In:
+  "Astronomical Techniques", ed. Hiltner, p. 184) to compute the airmass as a
+  function of zenith angle which is given in terms of its secons aSecZ.
+  It is supposedy more accurate than Young & Irvine (1967) and usable for zenith
+  angles below 85 degrees.
+ */
+/*----------------------------------------------------------------------------*/
+hdrl_value hdrl_get_airmass_hardie(hdrl_value	hvaSecZ)
+{
+	double aSecZ    = hvaSecZ.data,
+		   aSecZErr = hvaSecZ.error;
+
+	hdrl_value secm1 = {aSecZ - 1, aSecZErr};
+
+	hdrl_value airmass = {aSecZ - 0.0018167 * secm1.data
+							    - 0.002875  * secm1.data * secm1.data
+							    - 0.0008083 * secm1.data * secm1.data * secm1.data,
+                          aSecZErr +secm1.error * fabs( - 0.0018167
+									                    - 2. * 0.002875  * secm1.data
+									                    - 3. * 0.0008083 * secm1.data * secm1.data)};
+
+	return airmass;
+}
+
+/*----------------------------------------------------------------------------*/
+/**
+  @private
+  @brief   Compute airmass with Young and Irvine (1997) approximation.
+  @param   hvaSecZ   Secans of the zenith distance.
+  @return  The function returns the airmass.
+
+  The function uses the approximation given by Young and Irvine (Young A. T.,
+  Irvine W. M., 1967, Astron. J. 72, 945) to compute the airmass for a given
+  sec(z) aSecZ. This approximation takes into account atmosphere refraction and
+  curvature, but is in principle only valid at sea level.
+ */
+/*----------------------------------------------------------------------------*/
+hdrl_value hdrl_get_airmass_youngirvine(hdrl_value hvaSecZ)
+{
+	double aSecZ    = hvaSecZ.data,
+	       aSecZErr = hvaSecZ.error;
+
+	hdrl_value airmass = {aSecZ * (1. - 0.0012 * (pow(aSecZ, 2) - 1.)),
+	                      aSecZErr * fabs( (1. - 0.0012 * (pow(aSecZ, 2) - 1.)) - 2. * 0.0012 * pow(aSecZ, 2))};
+
+	return airmass;
+}
+
+/*----------------------------------------------------------------------------*/
+/**
+  @private
+  @brief   Compute airmass with Young (1994) approximation.
+  @param   aCosZt   The cosine of the true zenith distance.
+  @return  The function returns the airmass.
+
+  The function uses the approximation given by Young (Young A. T., 1994 ApOpt,
+  33, 1108) to compute the relative optical air mass as a function of true,
+  rather than refracted, zenith angle which is given in terms of its cosine
+  aCosZt.
+  It is supposedy more accurate than Young & Irvine (1967) but restrictions are
+  not known.
+ */
+/*----------------------------------------------------------------------------*/
+hdrl_value hdrl_get_airmass_young(hdrl_value hvaCosZt)
+{
+	double aCosZt    = hvaCosZt.data,
+		   aCosZtErr = hvaCosZt.error;
+
+	hdrl_value airmass = {(1.002432 * aCosZt * aCosZt + 0.148386 * aCosZt + 0.0096467)
+						/ (aCosZt * aCosZt * aCosZt + 0.149864 * aCosZt * aCosZt + 0.0102963 * aCosZt
+						   + 0.000303978),
+                 	    aCosZtErr * fabs( ( (2. * 1.002432 * aCosZt + 0.148386) * (aCosZt * aCosZt * aCosZt + 0.149864 * aCosZt * aCosZt + 0.0102963 * aCosZt + 0.000303978)
+									    - (3. * aCosZt * aCosZt + 2. * 0.149864 * aCosZt + 0.0102963) * (1.002432 * aCosZt * aCosZt + 0.148386 * aCosZt + 0.0096467)
+								     	) / pow(aCosZt * aCosZt * aCosZt + 0.149864 * aCosZt * aCosZt + 0.0102963 * aCosZt + 0.000303978, 2) )};
+
+	return airmass;
+}
+
+
 /**@}*/
