@@ -35,6 +35,7 @@
 
 #include <stdlib.h>
 #include <stdio.h>
+#include <math.h>
 
 /*----------------------------------------------------------------------------*/
 /**
@@ -43,6 +44,37 @@
  */
 /*----------------------------------------------------------------------------*/
 
+void test_create(void)
+{
+
+    /* create data initial data */
+    cpl_imagelist * data = cpl_imagelist_new();
+    cpl_imagelist * errs = cpl_imagelist_new();
+    cpl_image * img = cpl_image_new(64, 64, CPL_TYPE_DOUBLE);
+    cpl_image * err = cpl_image_new(64, 64, CPL_TYPE_DOUBLE);
+    cpl_image_add_scalar(img, 1.  );
+    cpl_image_add_scalar(err, 0.05);
+
+    /* create expected results (err / sqrt(nz) for mean) */
+    size_t n = 5;
+    for (size_t i = 0; i < n; i++) {
+        cpl_imagelist_set(data, cpl_image_duplicate(img), cpl_imagelist_get_size(data));
+        cpl_imagelist_set(errs, cpl_image_duplicate(err), cpl_imagelist_get_size(errs));
+    }
+
+    /* new hdrl_imagelist */
+	hdrl_imagelist * hl = hdrl_imagelist_create(data, errs);
+    cpl_test_eq(hdrl_imagelist_get_size(hl), n);
+    cpl_test_error(CPL_ERROR_NONE);
+
+	/* Clean up */
+	cpl_image_delete(img);
+	cpl_image_delete(err);
+	cpl_imagelist_delete(data);
+	cpl_imagelist_delete(errs);
+	hdrl_imagelist_delete(hl);
+}
+
 void test_get(void)
 {
     hdrl_imagelist * hl = NULL;
@@ -143,8 +175,27 @@ cpl_error_code test_iter(void)
     }
 
     for (cpl_size y = 1; y < ny + 1; y++) {
+
         hdrl_imagelist * lview = hdrl_imagelist_row_view(hl, y, y);
+
         hdrl_imagelist_add_scalar(lview, (hdrl_value){y, y});
+        hdrl_imagelist_sub_scalar(lview, (hdrl_value){y, y});
+
+        int isConsistent = hdrl_imagelist_is_consistent(lview);
+        cpl_msg_debug(cpl_func, "Imagelist consistent? %d", isConsistent);
+
+        /* Show the structure, the first time */
+        if(y == 1){
+
+        	hdrl_imagelist_dump_structure(lview, stdout);
+
+        	cpl_size llx = 1;
+        	cpl_size lly = 1;
+        	cpl_size urx = y;
+        	cpl_size ury = y;
+			hdrl_imagelist_dump_window(lview, llx, lly, urx, ury, stdout);
+        }
+
         hdrl_imagelist_delete(lview);
     }
 
@@ -190,13 +241,13 @@ cpl_error_code test_iter(void)
         cpl_test_eq(hdrl_imagelist_get_size(h), nz);
         hdrl_image * img = hdrl_imagelist_get(h, 0);
         cpl_test_eq(hdrl_image_get_size_x(img), nx);
-        if (ysum == 1 ||
-                ysum + hdrl_image_get_size_y(img) == ny - 1) {
+
+        if (ysum == 1) {
             cpl_test_leq(hdrl_image_get_size_y(img), yslice + overlap);
-        }
-        else {
+        } else {
             cpl_test_leq(hdrl_image_get_size_y(img), yslice + overlap * 2);
         }
+
         hdrl_image * m;
         cpl_image * c;
         hdrl_imagelist_collapse_mean(h, &m, &c);
@@ -238,6 +289,7 @@ int main(void)
 {
     cpl_test_init(PACKAGE_BUGREPORT, CPL_MSG_WARNING);
 
+    test_create();
     test_get();
     test_interface();
     test_iter();
