@@ -131,7 +131,8 @@ NACO_RECIPE_DEFINE(naco_spc_wavecal,
                    "The files listed in the Set Of Frames (sof-file) "
                    "must be tagged:\n" 
                    "NACO-raw-file.fits " NACO_SPC_LAMPWAVE_RAW "\n"
-                   "NACO-spectrum-model.fits " NACO_SPC_MODEL " .\n"
+                   "NACO-spectrum-model.fits " NACO_SPC_MODEL "\n"
+                   "NACO-arc-lines.fits " NACO_SPC_ARGON "\n"
                    "\n"
                    NACO_SPC_MAN_MODESPLIT "\n\n"
                    "Furthermore, each input frame must have a value of "
